/* genericups.c - support for generic contact-closure UPS models

   Copyright (C) 1999  Russell Kroll <rkroll@exploits.org>

   This program is free software; you can redistribute it and/or modify
   it under the terms of the GNU General Public License as published by
   the Free Software Foundation; either version 2 of the License, or
   (at your option) any later version.

   This program is distributed in the hope that it will be useful,
   but WITHOUT ANY WARRANTY; without even the implied warranty of
   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
   GNU General Public License for more details.

   You should have received a copy of the GNU General Public License
   along with this program; if not, write to the Free Software
   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
*/

#include <sys/ioctl.h>

#include "main.h"
#include "serial.h"
#include "genericups.h"

#define DRIVER_NAME	"Generic contact-closure UPS driver"
#define DRIVER_VERSION	"1.36"
#define TIOCM_OUT2 0x4000

/* driver description structure */
upsdrv_info_t upsdrv_info = {
	DRIVER_NAME,
	DRIVER_VERSION,
	"Russell Kroll <rkroll@exploits.org>",
	DRV_STABLE,
	{ NULL }
};

	static	int	upstype = -1;

#if defined(BUFFALO_FEATURE)
#define BUFFALO_MV78
#endif	/* BUFFALO_FEATURE */

#if defined(BUFFALO_MV78)
#undef  DRIVER_VERSION
#define DRIVER_VERSION	"1.36+BUFFALO"

#define BUFFALO_MV78_UPSPORT	"/dev/ttyS2"
static int buffalo_specific = 0;

static void buffalo_invert_signal(int *flags)
{
	int do_inv = TIOCM_CAR | TIOCM_RNG | TIOCM_DSR | TIOCM_DTR;

	if ((buffalo_specific == 0) ||
	    (strcmp(device_path, BUFFALO_MV78_UPSPORT) != 0)) {
		return;
	}

	upsdebugx(2, "%s: original flags = 0x%08x\n", __FUNCTION__, *flags);
	*flags ^= do_inv;
}

#define OMRON_UPS_SHUTDOWN_PROC	"/proc/buffalo/ups/omron_shutdown_enable"
#define UPS_SHUTDOWN_OP_OFF	0
static int ups_shutdown = UPS_SHUTDOWN_OP_OFF;	/* default: off */

static int buffalo_enable_kernel_support_shutdown_ups(int flag)
{
	FILE *fp;

	if (upstype != UPS_TYPE_OMRON) {
		/* this feature used for only OMRON */
		upslogx(LOG_ERR, "Kernel ups interface is only for OMRON, "
			"just ignored.\n");
		return -1;
	}

	fp = fopen(OMRON_UPS_SHUTDOWN_PROC, "r");
	if (fp == NULL) {
		upslogx(LOG_ERR, "Kernel don't support ups interface!\n");
		return -1;
	}
	fclose(fp);

	fp = fopen(OMRON_UPS_SHUTDOWN_PROC, "w");
	if (fp == NULL) {
		upslogx(LOG_ERR, "Can't open %s for write\n",
			OMRON_UPS_SHUTDOWN_PROC);
		return -1;
	}
	if (flag == UPS_SHUTDOWN_OP_OFF) {
		fprintf(fp, "off");
		upslogx(LOG_INFO, "disable kernel support to shutdown UPS.");
	}
	else {
		fprintf(fp, "on");
		upslogx(LOG_INFO, "enable kernel support to shutdown UPS.");
	}
	fclose(fp);

	return 0;
}
#endif	/* BUFFALO_MV78 */

static void parse_output_signals(const char *value, int *line)
{
	/* parse signals the serial port can output */

	*line = 0;

	if (strstr(value, "DTR") && !strstr(value, "-DTR")) {
		*line |= TIOCM_DTR;
	}

	if (strstr(value, "RTS") && !strstr(value, "-RTS")) {
		*line |= TIOCM_RTS;
	}

	if (strstr(value, "ST")) {
		*line |= TIOCM_ST;
	}

	if (strstr(value, "CTS")) {
		fatalx(EXIT_FAILURE, "Can't override output with CTS (not an output)");
	}

	if (strstr(value, "DCD")) {
		fatalx(EXIT_FAILURE, "Can't override output with DCD (not an output)");
	}

	if (strstr(value, "RNG")) {
		fatalx(EXIT_FAILURE, "Can't override output with RNG (not an output)");
	}

	if (strstr(value, "DSR")) {
		fatalx(EXIT_FAILURE, "Can't override output with DSR (not an output)");
	}
}

static void parse_input_signals(const char *value, int *line, int *val)
{
	/* parse signals the serial port can input */

	*line = 0;
	*val = 0;

	if (strstr(value, "CTS")) {
		*line |= TIOCM_CTS;

		if (!strstr(value, "-CTS")) {
			*val |= TIOCM_CTS;
		}
	}

	if (strstr(value, "DCD")) {
		*line |= TIOCM_CD;

		if (!strstr(value, "-DCD")) {
			*val |= TIOCM_CD;
		}
	}

	if (strstr(value, "RNG")) {
		*line |= TIOCM_RNG;

		if (!strstr(value, "-RNG")) {
			*val |= TIOCM_RNG;
		}
	}

	if (strstr(value, "DSR")) {
		*line |= TIOCM_DSR;

		if (!strstr(value, "-DSR")) {
			*val |= TIOCM_DSR;
		}
	}

	if (strstr(value, "DTR")) {
		fatalx(EXIT_FAILURE, "Can't override input with DTR (not an input)");
	}

	if (strstr(value, "RTS")) {
		fatalx(EXIT_FAILURE, "Can't override input with RTS (not an input)");
	}

	if (strstr(value, "ST")) {
		fatalx(EXIT_FAILURE, "Can't override input with ST (not an input)");
	}
}

void upsdrv_initinfo(void)
{
	char	*v;

	/* setup the basics */

	dstate_setinfo("ups.mfr", "%s", ((v = getval("mfr")) != NULL) ? v : upstab[upstype].mfr);
	dstate_setinfo("ups.model", "%s", ((v = getval("model")) != NULL) ? v : upstab[upstype].model);

	if ((v = getval("serial")) != NULL) {
		dstate_setinfo("ups.serial", "%s", v);
	}

	/*
	 User wants to override the input signal definitions. See also upsdrv_initups().
	 */
	if ((v = getval("OL")) != NULL) {
		parse_input_signals(v, &upstab[upstype].line_ol, &upstab[upstype].val_ol);
		upsdebugx(2, "parse_input_signals: OL overridden with %s\n", v);
	}

	if ((v = getval("LB")) != NULL) {
		parse_input_signals(v, &upstab[upstype].line_bl, &upstab[upstype].val_bl);
		upsdebugx(2, "parse_input_signals: LB overridden with %s\n", v);
	}
}

//Simple Shutdown Software
#define MAX_UPS_DATA_LENGTH		128
#define MAX_PORT_FAIL_LOOP		2
#define MAX_COMMAND_LENGTH		12

void DelayInUsec(int sec, int usecond)
{
	fd_set          fdpool;
	struct timeval  tm;

	FD_ZERO(&fdpool);
  
	tm.tv_sec  = sec;
	tm.tv_usec = usecond;
	select(10, &fdpool, NULL, NULL, &tm);
}

int WritePort(int fd, const char *pszData)
{
	char szCommand[MAX_COMMAND_LENGTH] = {0};
	int  iRet = 0;

	strcpy(szCommand, pszData);
	szCommand[strlen(pszData)] = 0x0d;
	szCommand[strlen(pszData)+1] = 0x0;

	upsdebugx(6, "write command = %s", szCommand);
	if((iRet = write(fd, szCommand, strlen(szCommand))) <= 0)
	{
		return 0;
	}
	upsdebugx(6, "write count = %d", iRet);
	return iRet;
}

int ReadPort(int fd, int iCount, char *pszUPSData)
{
	int iByteRead = 0;
	int iTotalByte = 0;
	int retryCount = 0;
	char *ptr = pszUPSData;
	fd_set readset;
	struct timeval tv;
	time_t tStartTime;

	memset(pszUPSData, 0, MAX_UPS_DATA_LENGTH);

	tStartTime = time(NULL);

	for(retryCount = 0; retryCount < 5; ++retryCount)
	{
		tv.tv_sec = 0;
		tv.tv_usec = 500000;
		FD_ZERO(&readset);	
		FD_SET(fd, &readset);

		if(select(fd+1, &readset, NULL, NULL, &tv) > 0)
		{
			if(FD_ISSET(fd, &readset))
			{
				iByteRead = read(fd, ptr+iTotalByte, iCount);
				upsdebugx(6, "iByteRead = %d", iByteRead);
				if(iByteRead == -1)
				{
					return 0;
				}
				iTotalByte += iByteRead;
				if((iTotalByte >= iCount) || strchr(ptr, '\n'))
				{
					break;
				}
			}
		}

		DelayInUsec(0, 500000);
	}

	upsdebugx(6, "read = %s", pszUPSData);
	upsdebugx(6, "read count = %d", iTotalByte);
	return iTotalByte;
}

int UPSCmdQ1(int fd, char *szData)
{
	int i;
	for(i=0; i<MAX_PORT_FAIL_LOOP; i++){
		if(WritePort(fd, "Q1") < 0){
			continue;
		}
		if(ReadPort(fd, MAX_UPS_DATA_LENGTH, szData) > 0)
		{
			upsdebugx(6, "UPSCmdQ1 read OK");
			return 1;
		}
	}
	return 0;
}
//Simple Shutdown Software

/* normal idle loop - keep up with the current state of the UPS */
void upsdrv_updateinfo(void)
{
	int	flags, ol, bl, ret;
	char 	szcmdBuf[MAX_UPS_DATA_LENGTH];

	if (upstype == UPS_TYPE_OMRON) {
		ret = UPSCmdQ1(upsfd, szcmdBuf);
		if (ret == 0) {
			upslogx(LOG_ERR, "Q1 command failed");
			ser_comm_fail("Status read failed");
			dstate_datastale();
			return;
		}
	}

	ret = ioctl(upsfd, TIOCMGET, &flags);

	if (ret != 0) {
		upslog_with_errno(LOG_INFO, "ioctl failed");
		ser_comm_fail("Status read failed");
		dstate_datastale();
		return;
	}

#if defined(BUFFALO_MV78)
	buffalo_invert_signal(&flags);
	upsdebugx(6, "flags=0x%08x", flags);
#endif	/* BUFFALO_MV78 */
	ol = ((flags & upstab[upstype].line_ol) == upstab[upstype].val_ol);
	bl = ((flags & upstab[upstype].line_bl) == upstab[upstype].val_bl);

	status_init();

	if (bl) {
		status_set("LB");	/* low battery */
	}

	if (ol) {
		status_set("OL");	/* on line */
	} else {
		status_set("OB");	/* on battery */
	}

	status_commit();

	upsdebugx(5, "ups.status: %s %s\n", ol ? "OL" : "OB", bl ? "BL" : "");

	ser_comm_good();
	dstate_dataok();
}

/* show all possible UPS types */
static void listtypes(void)
{
	int	i;

	printf("Valid UPS types:\n\n");

	for (i = 0; upstab[i].mfr != NULL; i++) {
		printf("%i: %s\n", i, upstab[i].desc);
	}
}

/* set the flags for this UPS type */
static void set_ups_type(void)
{
	int	i;

	if (!getval("upstype")) {
		fatalx(EXIT_FAILURE, "No upstype set - see help text / man page!");
	}

	upstype = atoi(getval("upstype"));

	for (i = 0; upstab[i].mfr != NULL; i++) {

		if (upstype == i) {
			upslogx(LOG_INFO, "UPS type: %s\n", upstab[i].desc);
			return;
		}
	}

	listtypes();

	fatalx(EXIT_FAILURE, "\nFatal error: unknown UPS type number");
}

/* power down the attached load immediately */
void upsdrv_shutdown(void)
{
	int	flags, ret;

	if (upstype == -1) {
		fatalx(EXIT_FAILURE, "No upstype set - see help text / man page!");
	}

	flags = upstab[upstype].line_sd;

	if (flags == -1) {
		fatalx(EXIT_FAILURE, "No shutdown command defined for this model!");
	}

	if (flags == TIOCM_ST) {

#ifndef HAVE_TCSENDBREAK
		fatalx(EXIT_FAILURE, "Need to send a BREAK, but don't have tcsendbreak!");
#endif

		ret = tcsendbreak(upsfd, 4901);

		if (ret != 0) {
			fatal_with_errno(EXIT_FAILURE, "tcsendbreak");
		}

		return;
	}

#if defined(BUFFALO_MV78)
	if (getval("upssd")) {
		ups_shutdown = atoi(getval("upssd"));
		upslogx(LOG_INFO, "kernel support to shutdown UPS = %d\n",
			ups_shutdown);
	}

	ret = buffalo_enable_kernel_support_shutdown_ups(ups_shutdown);
	if (ret != 0) {
		upslogx(LOG_ERR, "failed to set kernel support UPS shutdown\n");
	}

	buffalo_invert_signal(&flags);
#endif	/* BUFFALO_MV78 */
	ret = ioctl(upsfd, TIOCMSET, &flags);

	if (ret != 0) {
		fatal_with_errno(EXIT_FAILURE, "ioctl TIOCMSET");
	}

	if (getval("sdtime")) {
		int	sdtime;

		sdtime = strtol(getval("sdtime"), (char **) NULL, 10);

		upslogx(LOG_INFO, "Holding shutdown signal for %d seconds...\n",
			sdtime);

		sleep(sdtime);
	}
}

void upsdrv_help(void)
{
	listtypes();
}

void upsdrv_makevartable(void)
{
	addvar(VAR_VALUE, "upstype", "Set UPS type (required)");
	addvar(VAR_VALUE, "mfr", "Override manufacturer name");
	addvar(VAR_VALUE, "model", "Override model name");
	addvar(VAR_VALUE, "serial", "Specify the serial number");
	addvar(VAR_VALUE, "CP", "Override cable power setting");
	addvar(VAR_VALUE, "OL", "Override on line signal");
	addvar(VAR_VALUE, "LB", "Override low battery signal");
	addvar(VAR_VALUE, "SD", "Override shutdown setting");
	addvar(VAR_VALUE, "sdtime", "Hold time for shutdown value (seconds)");
#if defined(BUFFALO_MV78)
	addvar(VAR_VALUE, "buffalo", "enable BUFFALO specific feature");
	addvar(VAR_VALUE, "upssd", "use kernel support to shutdown UPS");
#endif	/* BUFFALO_MV78 */
}

void upsdrv_initups(void)
{
	struct termios	tio;
	char	*v;
#if defined(BUFFALO_MV78)
	int flags;
	int ret;
#endif	/* BUFFALO_MV78 */

	set_ups_type();

#if defined(BUFFALO_MV78)
	if (getval("buffalo")) {
		buffalo_specific = atoi(getval("buffalo"));
		upslogx(LOG_INFO, "Buffalo specific = %d\n", buffalo_specific);
	}
/*
	if (getval("upssd")) {
		ups_shutdown = atoi(getval("upssd"));
		upslogx(LOG_INFO, "kernel support to shutdown UPS = %d\n",
			ups_shutdown);
	}

	ret = buffalo_enable_kernel_support_shutdown_ups(ups_shutdown);
	if (ret != 0) {
		upslogx(LOG_ERR, "failed to set kernel support UPS shutdown\n");
	}
*/
#endif	/* BUFFALO_MV78 */

	upsfd = ser_open(device_path);

	if (tcgetattr(upsfd, &tio)) {
		fatal_with_errno(EXIT_FAILURE, "tcgetattr");
	}

	/* don't hang up on last close */
	tio.c_cflag &= ~HUPCL;
	if (upstype == UPS_TYPE_OMRON) {
		tio.c_cflag &= ~(CBAUD | PARENB | CSTOPB);
		tio.c_cflag |= (B2400 | CS8 | CLOCAL);
		tio.c_iflag |= (IXON | IXOFF | ICRNL);
		tio.c_iflag &= ~(ISTRIP | IGNCR);
		tio.c_oflag &= ~(ONOCR | ONLCR);
		tio.c_lflag &= ~(ICANON | ECHO);
		tio.c_lflag |= FLUSHO;
		tio.c_cc[VMIN] = 1;
		tio.c_cc[VTIME] = 10;
	}

	if (tcsetattr(upsfd, TCSANOW, &tio)) {
		fatal_with_errno(EXIT_FAILURE, "tcsetattr");
	}

	/*
	 See if the user wants to override the output signal definitions
	 this must be done here, since we might go to upsdrv_shutdown()
	 immediately. Input signal definition override is handled in
	 upsdrv_initinfo()
	 */
	if ((v = getval("CP")) != NULL) {
		parse_output_signals(v, &upstab[upstype].line_norm);
		upsdebugx(2, "parse_output_signals: CP overridden with %s\n", v);
	}

	if ((v = getval("SD")) != NULL) {
		parse_output_signals(v, &upstab[upstype].line_sd);
		upsdebugx(2, "parse_output_signals: SD overridden with %s\n", v);
	}

#if defined(BUFFALO_MV78)
	if (upstype == UPS_TYPE_OMRON) {
		flags = upstab[upstype].line_norm | TIOCM_OUT2;
	}else{
		flags = upstab[upstype].line_norm;
	}
	buffalo_invert_signal(&flags);
	if (ioctl(upsfd, TIOCMSET, &flags)) {
		fatal_with_errno(EXIT_FAILURE, "ioctl TIOCMSET");
	}
#else	/* BUFFALO_MV78 */
	if (ioctl(upsfd, TIOCMSET, &upstab[upstype].line_norm)) {
		fatal_with_errno(EXIT_FAILURE, "ioctl TIOCMSET");
	}
#endif	/* BUFFALO_MV78 */
}

void upsdrv_cleanup(void)
{
	ser_close(upsfd, device_path);
}

